/*
 * Program: SSL-WOPR Server
 * Source : wopr.cpp
 * Purpose: Implementation of WOPR_server class.
 *
 * Copyright (c) 2009, 2010 Oliver Mueller.
 * All rights reserved.
 * http://www.cogito-ergo-sum.org
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the author be held liable for any damages arising from
 * the use of this software.
 *
 */

// We are on z/OS Unix System Services?
#ifdef __IBMCPP__
#if __IBMCPP__ >= 20000
#define _XOPEN_SOURCE_EXTENDED 1
#define _OPEN_MSGQ_EXT
#include <strings.h>
#define WOPR_ZOS
#endif
#endif

#ifdef _MINIX
#define _SIZE_T
typedef long unsigned int size_t;
#define _SSIZE_T
typedef long int ssize_t;
#endif

#include <ctype.h>
#include <signal.h>
#include <string.h>

#include <sys/types.h>
#include <sys/wait.h>
#ifndef __hpux__
#include <sys/select.h>
#endif

#include <sys/time.h>
#include <unistd.h>

#include <openssl/err.h>

#include "woprexcept.h"
#include "wopr.h"

using namespace std;

#define BUFSIZE   8192

/* the static and class attribute stuff respectively */
bool WOPR_server::ssl_initialized = false;
bool WOPR_server::wopr_initialized = false;
const char *WOPR_server::identification = "\nWOPR V1.01\n";
const char *WOPR_server::prompt[2] = {"\nLOGON: ", // logon prompt
                                      "\n" // command prompt
};
unsigned WOPR_server::prompt_len[2] = {0, 0};
const char *WOPR_server::access_denied = "\n** IDENTIFICATION NOT RECOGNIZED **\n        ** ACCESS DENIED **\n";
const char *WOPR_server::already_running = "\n** GAME ROUTINE RUNNING **\n";
CmdNResp WOPR_server::cmd_n_resp[] = {
    { "LIST GAMES", 0,
     "\nFALKENS' MAZE\n"
     "BLACK JACK\n"
     "GIN RUMMY\n"
     "HEARTS\n"
     "BRIDGE\n"
     "CHECKERS\n"
     "CHESS\n"
     "POKER\n"
     "FIGHTER COMBAT\n"
     "GUERRILLA ENGAGEMENT\n"
     "DESERT WARFARE\n"
     "AIR-TO-GROUND ACTIONS\n"
     "THEATREWIDE TACTICAL WARFARE\n"
     "THEATREWIDE BIOTOXIC AND CHEMICAL WARFARE\n\n"
     "GLOBAL THERMONUCLEAR WAR\n", 0,
     0x00},
    { "FALKENS' MAZE", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "BLACK JACK", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "GIN RUMMY", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "HEARTS", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "BRIDGE", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "CHECKERS", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "CHESS", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "POKER", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "FIGHTER COMBAT", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "GUERRILLA ENGAGEMENT", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "DESERT WARFARE", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "AIR-TO-GROUND ACTIONS", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "THEATREWIDE TACTICAL WARFARE", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "THEATREWIDE BIOTOXIC AND CHEMICAL WARFARE", 0,
     WOPR_server::access_denied, 0,
     0x04},
    { "GLOBAL THERMONUCLEAR WAR", 0,
     WOPR_server::already_running, 0,
     0x04},
    { "TIC-TAC-TOE", 0,
     "\nA STRANGE GAME.\nTHE ONLY WINNING MOVE IS\nNOT TO PLAY.\n", 0,
     0x04},
    { "HELP LOGON", 0,
     "\nHELP NOT AVAILABLE\n", 0,
     0x00},
    { "HELP GAMES", 0,
     "\n'GAMES' REFERS TO MODELS, SIMULATIONS AND GAMES\n"
     "WHICH HAVE TACTICAL AND STRATEGIC APPLICATIONS.\n", 0,
     0x00},
    { "LOGOFF", 0,
     "\nGOODBYE\nCLOSING USER SESSION\n", 0,
     0x06},
    { "DISCONNECT", 0,
     "\nGOODBYE\nCLOSING CONNECTION\n\n", 0,
     0x16},
    { "JOSHUA", 0,
     "\nGREETINGS PROFESSOR FALKEN.\n", 0,
     0x09},
    { "MCKITTRICK", 0,
     "\nGREETINGS DOCTOR MCKITTRICK.\n", 0,
     0x09},
    { "", 0, "", 0, 0x00}
};
const char *WOPR_server::unknown_cmd =
        "\nUNKNOWN COMMAND\n\n";
unsigned WOPR_server::unknown_cmd_len = 0;
const char *WOPR_server::unknown_user =
        "IDENTIFICATION NOT RECOGNIZED BY SYSTEM\n"
        "--CONNECTION TERMINATED--\n\n";
unsigned WOPR_server::unknown_user_len = 0;

WOPR_server::WOPR_server(char *host, const char *cert, const char *key) throw (SSL_exception)
{
    ctx = NULL;
    abio = NULL;
    clno = 0L;
    if(!ssl_initialized) {
        SSL_library_init();
        SSL_load_error_strings();
        ERR_load_BIO_strings();
        ERR_load_SSL_strings();
        ssl_initialized = true;
    }
    if(!wopr_initialized) {
        for(int n = 0; cmd_n_resp[n].cmd[0] != '\0'; n++) {
            cmd_n_resp[n].cmd_len = strlen(cmd_n_resp[n].cmd);
            cmd_n_resp[n].resp_len = strlen(cmd_n_resp[n].resp);
        }
        prompt_len[0] = strlen(prompt[0]);
        prompt_len[1] = strlen(prompt[1]);
        unknown_cmd_len = strlen(unknown_cmd);
        unknown_user_len = strlen(unknown_user);
        wopr_initialized = true;
    }
    try {
        if((ctx = SSL_CTX_new(SSLv3_server_method())) == NULL ||
           SSL_CTX_use_certificate_file(ctx, cert, SSL_FILETYPE_PEM) != 1 ||
           SSL_CTX_use_PrivateKey_file(ctx, key, SSL_FILETYPE_PEM) != 1 ||
           (abio = BIO_new_accept(host)) == NULL) {
            ERR_error_string_n(ERR_get_error(), errmsgbuf, ERRSIZE);
            errmsgbuf[ERRSIZE - 1] = '\0';
            throw SSL_exception(errmsgbuf);
        }
    } catch(SSL_exception& se) {
        if(abio) {
            BIO_free_all(abio);
            abio = NULL;
        }
        if(ctx) {
            SSL_CTX_free(ctx);
            ctx = NULL;
        }
        throw se;
    }
}

WOPR_server::~WOPR_server()
{
    if(abio)
        BIO_free_all(abio);
    if(ctx)
        SSL_CTX_free(ctx);
}

void WOPR_server::run() throw (SSL_exception)
{
    if(!abio || !ctx)
        return;

    if(BIO_do_accept(abio) <= 0) {
        ERR_error_string_n(ERR_get_error(), errmsgbuf, ERRSIZE);
        errmsgbuf[ERRSIZE - 1] = '\0';
        throw SSL_exception(errmsgbuf);
    }

    fd_set fds;
    int afd = BIO_get_fd(abio, NULL);
    struct timeval timeout;

    while(1) {
        FD_ZERO(&fds);
        FD_SET(afd, &fds);
        timeout.tv_sec = 0L;
        timeout.tv_usec = 100L;
        select(afd + 1, &fds, (fd_set *) 0, (fd_set *) 0, &timeout);
        if(FD_ISSET(afd, &fds) && BIO_do_accept(abio) > 0) {
            clno++;
            pid_t pid = fork();
            client = BIO_pop(abio);
            switch(pid) {
                case 0:
                    ssl = SSL_new(ctx);
                    if(!ssl) {
                        print_client_error("Error getting SSL context");
                        exit(1);
                    }
                    SSL_set_accept_state(ssl);
                    SSL_set_bio(ssl, client, client);
                    if(SSL_accept(ssl) <= 0) {
                        print_client_error("Error at connection accept");
                        ERR_print_errors_fp(stderr);
                        exit(1);
                    } else {
                        try {
                            handle_client();
                        } catch(exception& e) {
                            print_client_error(e.what());
                        }
                        if(SSL_get_shutdown(ssl) & SSL_RECEIVED_SHUTDOWN)
                            SSL_shutdown(ssl);
                        else
                            SSL_clear(ssl);
                        cout << "Client no. " << clno << " closed SSL connection." << endl;
                        SSL_free(ssl);
                        ERR_remove_state(0);
                    }
                    cout << "Client no. " << clno << " terminated." << endl;
                    exit(0);
                    break;
                case -1:
                    perror("Fork error");
                    break;
                default:
                    // Close socket of BIO client on parent process.
                    // Otherwise our server will run out of open files
                    // (see ulimit), because it will keep all sockets open.
                    close(BIO_get_fd(client, NULL));
                    cout << "Client no. " << clno << " forked. PID " << pid << endl;
            }
        }
        waitpid(0, NULL, WNOHANG);
    }
}

void WOPR_server::handle_client() throw (Overflow_exception)
{
    fd_set rfds, wfds;
    int cfd = BIO_get_fd(client, NULL), maxfd = cfd + 1, r, wbuf_len = 0, wbuf_offset = 0, n, resp_len;
    char rbuf[BUFSIZE + 1], wbuf[BUFSIZE + 1];
    const char *resp;
    bool shutdown = false;
    short currpmpt = 0;
    bool nocmd;

#ifdef WOPR_ZOS
    char ascii_resp[BUFSIZE + 1];
#endif

    rbuf[BUFSIZE] = '\0';
    wbuf[BUFSIZE] = '\0';

    /*
     * To switch do non-blocking mode is not necessary,
     * according to BIO_s_connect(3) the BIO is already
     * in non-blocking mode after establishing of a
     * connection:
     * "BIO_set_nbio() sets the non blocking I/O flag to n.
     *  If n is zero then blocking I/O is set. If n is 1
     *  then non blocking I/O is set. Blocking I/O is
     *  the default. The call to BIO_set_nbio() should
     *  be made before the connection is established
     *  because non blocking I/O is set during the
     *  connect process."
     */

    strcpy(wbuf, identification);
    strcat(wbuf, prompt[currpmpt]);

#ifdef WOPR_ZOS
    // convert welcome screen from EBCDIC to ASCII
    __etoa(wbuf);
#endif

    wbuf_len = strlen(wbuf);

    for(;;) {
        FD_ZERO(&rfds);
        FD_ZERO(&wfds);
        FD_SET(cfd, &rfds);
        if(wbuf_len)
            FD_SET(cfd, &wfds);

        if((r = select(maxfd, &rfds, &wfds, (fd_set*) 0, (struct timeval*) 0)) == 0)
            continue;

        if(FD_ISSET(cfd, &rfds)) {
            do {
                r = SSL_read(ssl, rbuf, BUFSIZE);
                switch(SSL_get_error(ssl, r)) {
                    case SSL_ERROR_NONE:
                        rbuf[r] = '\0';

#ifdef WOPR_ZOS
                        // convert read data from ASCII to EBCDIC
                        __atoe(rbuf);
#endif

                        trim_all(rbuf);

                        if(rbuf[0] == '\n' || rbuf[0] == '\r') // empty line => skip
                            continue;

                        /* process received commands */
                        nocmd = true;
                        for(n = 0; cmd_n_resp[n].cmd[0] != '\0'; n++) {
                            if((rbuf[cmd_n_resp[n].cmd_len] == '\r' || rbuf[cmd_n_resp[n].cmd_len] == '\n') && strncasecmp(rbuf, cmd_n_resp[n].cmd, cmd_n_resp[n].cmd_len) == 0) {
                                if(currpmpt == 0 && (cmd_n_resp[n].status & STA_LGNOLY)) { // only at logged on state
                                    cout << "Only at logged on state" << endl;
                                    break;
                                }
                                if(currpmpt == 1 && (cmd_n_resp[n].status & STA_LGFOLY)) { // only at logged off state
                                    cout << "Only at logged off state" << endl;
                                    break;
                                }
                                nocmd = false;
                                if(cmd_n_resp[n].status & STA_LOGSET) // set logged on state?
                                    currpmpt = 1;
                                if(cmd_n_resp[n].status & STA_LOGUNS) // set logged off state?
                                    currpmpt = 0;
                                if(cmd_n_resp[n].status & STA_SHTDWN) // set shutdown?
                                    shutdown = true;
                                break;
                            }
                        }

                        /* print status */
                        print_client_info("BEGIN status of received input");
                        cout << "Received input: " << rbuf;
                        if(nocmd)
                            cout << "No command found." << endl;
                        else
                            cout << "Command found: " << cmd_n_resp[n].cmd << endl;
                        cout << "Logged on? " << (currpmpt == 1 ? "YES" : "NO") << endl
                                << "SSL shutdown planned? " << (shutdown ? "YES" : "NO") << endl;
                        print_client_info("END status of received input");

                        if(nocmd) {
                            if(currpmpt == 0) {
                                resp = unknown_user;
                                resp_len = unknown_user_len;
                                shutdown = true;
                            } else {
                                resp = unknown_cmd;
                                resp_len = unknown_cmd_len;
                            }
                        } else {
                            resp = cmd_n_resp[n].resp;
                            resp_len = cmd_n_resp[n].resp_len;
                        }

                        if(wbuf_len + resp_len > BUFSIZE)
                            throw Overflow_exception();

#ifdef WOPR_ZOS
                        // convert response from EBCDIC to ASCII
                        strcpy(ascii_resp, resp);
                        __etoa(ascii_resp);
                        strcat(wbuf, ascii_resp);
#else
                        strcat(wbuf, resp);
#endif
                        wbuf_len += resp_len;

                        if(!shutdown) {
#ifdef WOPR_ZOS
                            // convert prompt from EBCDIC to ASCII
                            strcpy(ascii_resp, prompt[currpmpt]);
                            __etoa(ascii_resp);
                            strcat(wbuf, ascii_resp);
#else
                            strcat(wbuf, prompt[currpmpt]);
#endif
                            wbuf_len += prompt_len[currpmpt];
                        }

                        break;
                    case SSL_ERROR_ZERO_RETURN:
                        return;
                    default:
                        print_client_error("SSL read problem");
                        return;
                }
            } while(SSL_pending(ssl));

        }

        if(wbuf_len && (FD_ISSET(cfd, &wfds) || shutdown)) {
            r = SSL_write(ssl, wbuf + wbuf_offset, wbuf_len);
            switch(SSL_get_error(ssl, r)) {
                case SSL_ERROR_NONE:
                    wbuf_len -= r;
                    if(wbuf_len <= 0) {
                        wbuf[0] = '\0';
                        wbuf_offset = 0;
                    } else
                        wbuf_offset += r;
                    break;
                case SSL_ERROR_WANT_WRITE:
                    break;
                default:
                    print_client_error("SSL write problem");
                    return;
            }
        }

        if(shutdown)
            return;
    }
}

void WOPR_server::trim_all(char *s)
{
    long n, m;
    bool first_space = true;

    if(s[0] == ' ' || s[0] == '\t') {
        first_space = false;
        n = 1;
    } else
        n = 0;

    for(m = 0; s[n] != '\0'; n++, m++) {
        if(s[n] == '\t' || s[n] == ' ') {
            if(first_space)
                first_space = false;
            else
                m--;
        } else if((s[n] == '\r' || s[n] == '\n') && m > 0 && (s[n - 1] == ' ' || s[n - 1] == '\t')) {
            m--;
        } else
            first_space = true;
        s[m] = (s[n] == '\t' ? ' ' : s[n]);
    }
    s[m] = '\0';
}
